--
--  AY3819X.vhd
--
--  Top entity of AY3819X.
--
--        Copyright (C)2001 - 2002 SEILEBOST
--                   All rights reserved.
--
-- $Id: AY3819X.vhd, v0.1 2002/30/04 00:00:00 SEILEBOST $
--
-- TODO :
--   Many verification !!
-- Remark :

library IEEE;
library UNISIM;
use IEEE.STD_LOGIC_1164.ALL;
use IEEE.numeric_STD.all;
use IEEE.STD_LOGIC_UNSIGNED.ALL; 
use UNISIM.Vcomponents.ALL; -- for IOBUF and OBUF

entity AY3819X is
    Port ( D       : inout  std_logic_vector(7 downto 0); -- DATA from/to CPU
           ADR     : in     std_logic_vector(3 downto 0); -- Address from CPU
           RESET   : in     std_logic;                    -- RESET
           CLOCK   : in     std_logic;                    -- Master Clock 16 MHz
           CSFSP   : in     std_logic;                    -- FSP Chip Select
           RW      : in     std_logic;                    -- 0 => READ / 1 => WRITE
           AnalogA : out    std_logic;                    -- Analog output for TONE A
           AnalogB : out    std_logic;                    -- Analog output for TONE B
           AnalogC : out    std_logic );                  -- Analog output for TONE C
end AY3819X;


architecture Behavioral of AY3819X is

 SIGNAL BUS_CS : std_logic_vector(15 downto 0); -- Select the different module when Read / Write Register

 -- Create register 
 SIGNAL R0    : std_logic_vector(7  downto 0);  -- Tone generator frequency Fine Tune channel A
 SIGNAL R1    : std_logic_vector(3  downto 0);  -- Tone generator frequency Coarse Tune channel A
 SIGNAL R2    : std_logic_vector(7  downto 0);  -- Tone generator frequency Fine Tune  channel B
 SIGNAL R3    : std_logic_vector(3  downto 0);  -- Tone generator frequency Coarse Tune channel B
 SIGNAL R4    : std_logic_vector(7  downto 0);  -- Tone generator frequency Fine Tune channel C
 SIGNAL R5    : std_logic_vector(3  downto 0);  -- Tone generator frequency Coarse Tune channel B
 SIGNAL R6    : std_logic_vector(7  downto 0);  -- Noise generator frequency
 SIGNAL R7    : std_logic_vector(7  downto 0);  -- Mixer Control I/O Enable
 SIGNAL R8    : std_logic_vector(7  downto 0);  -- Amplitude control channel A
 SIGNAL R9    : std_logic_vector(7  downto 0);  -- Amplitude control channel B
 SIGNAL R10   : std_logic_vector(7  downto 0);  -- Amplitude control channel C
 SIGNAL R11   : std_logic_vector(7  downto 0);  -- Envelope period control fine tune 
 SIGNAL R12   : std_logic_vector(3  downto 0);  -- Envelope period control coarse tune
 SIGNAL R13   : std_logic_vector(7  downto 0);  -- Envelope shape/cycle control

 SIGNAL CLK_A        : std_logic; -- CLOCK TONE VOICE A
 SIGNAL CLK_B        : std_logic; -- CLOCK TONE VOICE B
 SIGNAL CLK_C        : std_logic; -- CLOCK TONE VOICE C 
 SIGNAL CLK_TONE_A   : std_logic; -- CLOCK TONE VOICE A +/- CLOCK NOISE
 SIGNAL CLK_TONE_B   : std_logic; -- CLOCK TONE VOICE B +/- CLOCK NOISE
 SIGNAL CLK_TONE_C   : std_logic; -- CLOCK TONE VOICE C +/- CLOCK NOISE
 SIGNAL CLK_E        : std_logic; -- CLOCK Envelope Generator
 SIGNAL CLK_N        : std_logic; -- CLOCK From Noise Generator
 SIGNAL CLK_16       : std_logic; -- CLOCK (1 /16 MHz) 

 SIGNAL OUT_AMPL_E   : std_logic_vector(3 downto 0); -- Amplitude of signal from Envelope generator

 SIGNAL IAnalogA     : std_logic; -- FOR IOPAD, exit from DAC VOICE A
 SIGNAL IAnalogB     : std_logic; -- FOR IOPAD, exit from DAC VOICE B
 SIGNAL IAnalogC     : std_logic; -- FOR IOPAD, exit from DAC VOICE C

 SIGNAL fLD_TONE_A   : std_logic; -- Flag to modify frenquency of TONE A
 SIGNAL fLD_TONE_B   : std_logic; -- Flag to modify frenquency of TONE B
 SIGNAL fLD_TONE_C   : std_logic; -- Flag to modify frenquency of TONE C
 SIGNAL fLD_ENV      : std_logic; -- Flag to modify frenquency of Envelope
 SIGNAL fLD_NOISE    : std_logic; -- Flag to modify frenquency of Noise
 SIGNAL fLD_TONE_ENV : std_logic; -- Flag to modify frenquency of TONE Envelope

 -- To generator a square signal
 COMPONENT TONE_GENERATOR  PORT ( CLK         : in     std_logic;
                                  RST         : in     std_logic; 
                                  LD_TONE     : in     std_logic;
                                  FRQ_TONE_H  : in     std_logic_vector(3 downto 0);
                                  FRQ_TONE_L  : in     std_logic_vector(7 downto 0);
                                  OUT_TONE    : inout  std_logic );
 END COMPONENT;

 -- To generate a random signal
 COMPONENT NOISE_GENERATOR  PORT ( CLK          : in     std_logic;
                                   RST          : in     std_logic;
                                   LD_NOISE     : in     std_logic;
                                   DATA         : in     std_logic_vector(4 downto 0);
                                   CLK_N        : out    std_logic );
 END COMPONENT;

 -- TO GENERATE A 1/16 MHz clock for tone and envelope generator
 COMPONENT GEN_CLK  PORT ( CLK      : in  std_logic;
                           RST      : in  std_logic;
                           CLK_16   : out std_logic );
 END COMPONENT;

 -- TO MIX TONE, NOISE and ENVELOPE GENERATOR
 COMPONENT MIXER    PORT ( CLK          : in     std_logic;
                           RST          : in     std_logic;
                           IN_A         : in     std_logic;
                           IN_B         : in     std_logic;
                           IN_C         : in     std_logic;
                           IN_NOISE     : in     std_logic;
                           DATA         : in     std_logic_vector(5 downto 0);
                           OUT_A        : out    std_logic;
                           OUT_B        : out    std_logic;
                           OUT_C        : out    std_logic );                           
 END COMPONENT; 

 -- To generate enveloppe of a tone (A, B, C)
 COMPONENT GEN_ENV  PORT ( CLK_ENV      : in     std_logic;
                           DATA         : in     std_logic_vector(3 downto 0);
                           LD_ENV       : in     std_logic;
                           RST          : in     std_logic;
                           OUT_DATA     : inout  std_logic_vector(3 downto 0));
 END COMPONENT;

  -- To manage amplitude of tone (A, B, C)
 COMPONENT MANAGE_AMPLITUDE  PORT ( CLK           : in   std_logic;
                                    CLK_DAC       : in   std_logic;
                                    CLK_TONE      : in   std_logic;
                                    RST           : in   std_logic;
                                    AMPLITUDE     : in   std_logic_vector(4 downto 0);
                                    AMPLITUDE_E   : in   std_logic_vector(3 downto 0);
                                    OUT_DAC       : out  std_logic );
 END COMPONENT;

 COMPONENT IOBUF_F_12 port (   O : out   std_logic;
                              IO : inout std_logic;
                               I : in    std_logic;
                               T : in    std_logic  );
 END COMPONENT;

 COMPONENT OBUF_F_12 port (   O : out   std_logic;
                             IO : inout std_logic;
                              I : in    std_logic;
                              T : in    std_logic  );
 END COMPONENT;

 component OBUF_F_24
 port (
   I : in std_logic;
   O : out std_logic );
 end component;

BEGIN

U_TRAIT : PROCESS(CLOCK, RESET, ADR, D)
BEGIN
  IF (RESET = '1') THEN
     R0       <= "00000000"; 
     R1       <= "0000"; 
     R2       <= "00000000"; 
     R3       <= "0000"; 
     R4       <= "00000000"; 
     R5       <= "0000"; 
     R6       <= "00000000"; 
     R7       <= "00000000"; 
     R8       <= "00000000"; 
     R9       <= "00000000"; 
     R10      <= "00000000";      
     R11      <= "00000000"; 
     R12      <= "0000"; 
     R13      <= "00000000"; 
     D        <= "ZZZZZZZZ";  
  ELSIF rising_edge(CLOCK) THEN -- edge clock
    fLD_TONE_A   <= '0';
    fLD_TONE_B   <= '0';
    fLD_TONE_C   <= '0';
    fLD_ENV      <= '0';
    fLD_NOISE    <= '0';
    fLD_TONE_ENV <= '0';
    IF (CSFSP = '1') THEN       
       IF (RW = '0') THEN -- READ
          CASE ADR is
           WHEN "0000" => D <= R0;
           WHEN "0001" => D <= "0000" & R1;
           WHEN "0010" => D <= R2;
           WHEN "0011" => D <= "0000" & R3;
           WHEN "0100" => D <= R4;
           WHEN "0101" => D <= "0000" & R5;
           WHEN "0110" => D <= R6;
           WHEN "0111" => D <= R7;
           WHEN "1000" => D <= R8;
           WHEN "1001" => D <= R9;
           WHEN "1010" => D <= R10;
           WHEN "1011" => D <= R11;
           WHEN "1100" => D <= "0000" & R12;
           WHEN "1101" => D <= R13;
           WHEN OTHERS => D <= "ZZZZZZZZ";             
          END CASE;
       ELSE -- WRITE
          CASE ADR is
           WHEN "0000" => R0  <= D;
           WHEN "0001" => R1  <= D(3 downto 0); fLD_TONE_A <= '1';
           WHEN "0010" => R2  <= D;
           WHEN "0011" => R3  <= D(3 downto 0); fLD_TONE_B <= '1';
           WHEN "0100" => R4  <= D;
           WHEN "0101" => R5  <= D(3 downto 0); fLD_TONE_C <= '1';
           WHEN "0110" => R6  <= D;             fLD_NOISE  <= '1';
           WHEN "0111" => R7  <= D;
           WHEN "1000" => R8  <= D;
           WHEN "1001" => R9  <= D;
           WHEN "1010" => R10 <= D;
           WHEN "1011" => R11 <= D; 
           WHEN "1100" => R12 <= D(3 downto 0); fLD_TONE_ENV <= '1';
           WHEN "1101" => R13 <= D;             fLD_ENV      <= '1';
           WHEN OTHERS => D   <= "ZZZZZZZZ";             
          END CASE;
       END IF;
     END IF;
  END IF;
  
end PROCESS;

-- Instantiation of sub_level modules

UTONE_A : TONE_GENERATOR PORT MAP( CLK         => CLK_16,
                                   RST         => RESET,
                                   LD_TONE     => fLD_TONE_A,
                                   FRQ_TONE_H  => R1,
                                   FRQ_TONE_L  => R0,
                                   OUT_TONE    => CLK_A);

UTONE_B : TONE_GENERATOR PORT MAP( CLK         => CLK_16,
                                   RST         => RESET,
                                   LD_TONE     => fLD_TONE_B,
                                   FRQ_TONE_H  => R3,
                                   FRQ_TONE_L  => R2,
                                   OUT_TONE    => CLK_B);

UTONE_C : TONE_GENERATOR PORT MAP( CLK         => CLK_16,
                                   RST         => RESET,
                                   LD_TONE     => fLD_TONE_C,
                                   FRQ_TONE_H  => R5,
                                   FRQ_TONE_L  => R4,
                                   OUT_TONE    => CLK_C);

UTONE_NOISE : NOISE_GENERATOR PORT MAP( CLK      => CLK_16,
                                        RST      => RESET,
                                        LD_NOISE => fLD_NOISE,
                                        DATA     => R6(4 downto 0),
                                        CLK_N    => CLK_N);

UTONE_ENV : TONE_GENERATOR PORT MAP( CLK         => CLK_16,
                                     RST         => RESET,
                                     LD_TONE     => fLD_TONE_ENV,
                                     FRQ_TONE_H  => R12,
                                     FRQ_TONE_L  => R11,
                                     OUT_TONE    => CLK_E);

UCLK : GEN_CLK PORT MAP( CLK      => CLOCK, 
                         RST      => RESET,
                         CLK_16   => CLK_16 );

UMIXER : MIXER PORT MAP ( CLK          => CLOCK,
                          RST          => RESET,
                          IN_A         => CLK_A,
                          IN_B         => CLK_B,
                          IN_C         => CLK_C,
                          IN_NOISE     => CLK_N,
                          DATA         => R7(5 downto 0),
                          OUT_A        => CLK_TONE_A,
                          OUT_B        => CLK_TONE_B,
                          OUT_C        => CLK_TONE_C);

UGenEnv : GEN_ENV PORT MAP( CLK_ENV      => CLK_E,
                            DATA         => R13(3 downto 0),
                            LD_ENV       => fLD_ENV,
                            RST          => RESET,
                            OUT_DATA     => OUT_AMPL_E);

UManAmpA : MANAGE_AMPLITUDE PORT MAP ( CLK           => CLOCK,
                                       CLK_DAC       => CLOCK,
                                       CLK_TONE      => CLK_TONE_A,
                                       RST           => RESET,
                                       AMPLITUDE     => R8(4 downto 0),
                                       AMPLITUDE_E   => OUT_AMPL_E(3 downto 0),
                                       OUT_DAC       => IAnalogA );                                       

UManAmpB : MANAGE_AMPLITUDE PORT MAP ( CLK           => CLOCK,
                                       CLK_DAC       => CLOCK,
                                       CLK_TONE      => CLK_TONE_B,
                                       RST           => RESET,
                                       AMPLITUDE     => R9(4 downto 0),
                                       AMPLITUDE_E   => OUT_AMPL_E(3 downto 0),
                                       OUT_DAC       => IAnalogB );                                          

UManAmpC : MANAGE_AMPLITUDE PORT MAP ( CLK           => CLOCK,
                                       CLK_DAC       => CLOCK,
                                       CLK_TONE      => CLK_TONE_C,
                                       RST           => RESET,
                                       AMPLITUDE     => R10(4 downto 0),
                                       AMPLITUDE_E   => OUT_AMPL_E(3 downto 0),
                                       OUT_DAC       => IAnalogC );      
                                  

PAD_ANALOGA : OBUF_F_24 port map( I => IAnalogA, O => AnalogA);
PAD_ANALOGB : OBUF_F_24 port map( I => IAnalogB, O => AnalogB);
PAD_ANALOGC : OBUF_F_24 port map( I => IAnalogC, O => AnalogC);

end Behavioral;
